; file: flash_copy.s
; loader for Altera-compatible flow: user software is flash-unaware, linked for RAM, gets copied there

; assemble with e.g.:
; nios-elf-as -I ./platform/nios/sdk/inc --defsym __nios32__=1 -m32 ./platform/nios/sdk/startup/flash_copy.s -o obj/flash_copy.s.o
; link with e.g.:
; nios-elf-ld -Ttext=0x40000 obj/flash_copy.s.o -o obj/flash_copy.out
; check with:
; nios-elf-objdump -D obj/flash_copy.out
; make srec with:
; nios-elf-objcopy -O srec obj/flash_copy.out loader.flash

.include "nios.s"         ; macros and memory definition
.include "filesize.s"     ; generated header file, containing 'filesz' definition for the particular compilation

.global _start

_start:
	.org 0x00000          ; this will be 0x40000 after linking
	BR continue           ; skip over the ID string location, this also leaves us some room for parameters
	NOP                   ; delay slot
length:
	; file size from the included file, for the particular code to be loaded
	.int filesz           ; length of to-be-copied data, "in the clear", so it can be patched in easily

	.org 0x0000C          ; hard-pin the address instead of padding the above, to be safe
	.ascii "Nios"         ; GERMS monitor checks for this, at location 0x4000C (after linking)

continue:
	MOVIP %l0, 0x040100   ; %l0 is source (flash) starting adddress
	MOVIP %l1, na_ram     ; %l1 is destination (SRAM) address start location, defined in nios_macros.s
	MOVIP %l2, 128        ; %l2 is the address increment per loop, unrolled chunk
	MOVIA %l4, length     ; get the (linked) address of label "length"
	LD %l4, [%l4]         ; %l4 is the copy length (in bytes)

	; round the length up to 32 bit boundary
	; not really necessary for the bare copy loop, but for the calculated jump
	MOVI %l5, 3           ; rounding (alignment-1)
	ADD %l4, %l5          ; add rounding value
	NOT %l5               ; make a mask from it, to clear lower bits
	AND %l4, %l5          ; %l4 is rounded up to 32bit alignment

	; prepare loop check
	MOV %l3, %l1          ; copy the destination address to %l3
	ADD %l3, %l4          ; %l3 is now the destination end address (comparison value)


	; nasty optimization trick like only possible in assembler:
	; The unrolled copy block copies big chunks of 128 bytes.
	; To circumvent that granularity, the first iteration is partial, 
	; doing the "unaligned" part by a calculated jump into the code block.
	; We need to backwards-offset the source and destination address beforehand,
	; such that the jumped instructions work on the start address, despite the coded offset.
	; Conveniently, one load/store instruction pair is 4 bytes,
	; same as the amount of copied data, so we can use data size for code jump offset.
	
	MOV %l5, %l2          ; copy loop increment into %l5
	DEC %l5               ; make an AND mask from it, to keep relevant lower bits
	MOV %l6, %l4          ; copy rounded length into %l6
	NEG %l6               ; to get 128-len
	AND %l6, %l5          ; %l6 now is 128 - (length modulo 128), the address correction value
	SUB %l0, %l6          ; correct the source address
	SUB %l1, %l6          ; correct the destination address

	MOVIA %g0, loop       ; load the address of the copy block into %g0
	ADD %g0, %l6          ; add the correction as jump offset
	LSRI %g0, 1           ; divide by 2, convert to a jump argument
	JMP %g0               ; jump to proper place of the copy block
	MOV %g1, %l1          ; delay slot: save the destination into %g1, for later start
	
	; In doing above, on average we save half a copy block, 
	; 32 instructions and 32 memory bus cycles, by spending an extra 16 instructions.
	; The whole thing is not really worth it here, but done as an exercise.

	; copy from %l0 to %l1, increment by %l2 each iteration, until %l0 reaches %l3
loop:                     ; unrolled up to maximum reach of indexed address offset
	LDP %g0, [%l0, 0]     ; get current word
	STP [%l1, 0], %g0     ; store word
	LDP %g0, [%l0, 1]     ; get current word
	STP [%l1, 1], %g0     ; store word
	LDP %g0, [%l0, 2]     ; get current word
	STP [%l1, 2], %g0     ; store word
	LDP %g0, [%l0, 3]     ; get current word
	STP [%l1, 3], %g0     ; store word
	LDP %g0, [%l0, 4]     ; get current word
	STP [%l1, 4], %g0     ; store word
	LDP %g0, [%l0, 5]     ; get current word
	STP [%l1, 5], %g0     ; store word
	LDP %g0, [%l0, 6]     ; get current word
	STP [%l1, 6], %g0     ; store word
	LDP %g0, [%l0, 7]     ; get current word
	STP [%l1, 7], %g0     ; store word
	LDP %g0, [%l0, 8]     ; get current word
	STP [%l1, 8], %g0     ; store word
	LDP %g0, [%l0, 9]     ; get current word
	STP [%l1, 9], %g0     ; store word
	LDP %g0, [%l0, 10]    ; get current word
	STP [%l1, 10], %g0    ; store word
	LDP %g0, [%l0, 11]    ; get current word
	STP [%l1, 11], %g0    ; store word
	LDP %g0, [%l0, 12]    ; get current word
	STP [%l1, 12], %g0    ; store word
	LDP %g0, [%l0, 13]    ; get current word
	STP [%l1, 13], %g0    ; store word
	LDP %g0, [%l0, 14]    ; get current word
	STP [%l1, 14], %g0    ; store word
	LDP %g0, [%l0, 15]    ; get current word
	STP [%l1, 15], %g0    ; store word
	LDP %g0, [%l0, 16]    ; get current word
	STP [%l1, 16], %g0    ; store word
	LDP %g0, [%l0, 17]    ; get current word
	STP [%l1, 17], %g0    ; store word
	LDP %g0, [%l0, 18]    ; get current word
	STP [%l1, 18], %g0    ; store word
	LDP %g0, [%l0, 19]    ; get current word
	STP [%l1, 19], %g0    ; store word
	LDP %g0, [%l0, 20]    ; get current word
	STP [%l1, 20], %g0    ; store word
	LDP %g0, [%l0, 21]    ; get current word
	STP [%l1, 21], %g0    ; store word
	LDP %g0, [%l0, 22]    ; get current word
	STP [%l1, 22], %g0    ; store word
	LDP %g0, [%l0, 23]    ; get current word
	STP [%l1, 23], %g0    ; store word
	LDP %g0, [%l0, 24]    ; get current word
	STP [%l1, 24], %g0    ; store word
	LDP %g0, [%l0, 25]    ; get current word
	STP [%l1, 25], %g0    ; store word
	LDP %g0, [%l0, 26]    ; get current word
	STP [%l1, 26], %g0    ; store word
	LDP %g0, [%l0, 27]    ; get current word
	STP [%l1, 27], %g0    ; store word
	LDP %g0, [%l0, 28]    ; get current word
	STP [%l1, 28], %g0    ; store word
	LDP %g0, [%l0, 29]    ; get current word
	STP [%l1, 29], %g0    ; store word
	LDP %g0, [%l0, 30]    ; get current word
	STP [%l1, 30], %g0    ; store word
	LDP %g0, [%l0, 31]    ; get current word
	STP [%l1, 31], %g0    ; store word
	ADD %l1, %l2          ; bump destination to next block
	CMP %l1, %l3          ; are we there yet?
	BLT loop              ; not yet: continue the loop
	ADD %l0, %l2          ; delay slot: bump source to next block

	; jump to the destination address, saved in %g1
	LSRI %g1, 1           ; divide by 2, convert to a jump argument
	JMP %g1               ; jump to RAM start
	NOP                   ; delay slot

; End of file

